﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net.Http;
using System.Runtime.InteropServices;
using System.Security.Cryptography;
using System.Text;
using System.Threading.Tasks;
using Newtonsoft.Json;

namespace VitriumApiSampleCode
{
   public class ApiRequest
   {
      private const string SessionTokenKey = "X-VITR-SESSION-TOKEN";
      private const string AccountTokenKey = "X-VITR-ACCOUNT-TOKEN";

      private string _baseUrl;
      private string _sessionToken;
      private string _accountToken;

      public ApiRequest(string baseUrl, string username, string password, string accountId)
      {
         _baseUrl = baseUrl;
         Login(username, password, accountId);
      }

      private void SetHeaders(HttpClient client)
      {
         if (_sessionToken != null && _accountToken != null)
         {
            client.DefaultRequestHeaders.Add(SessionTokenKey, _sessionToken);
            client.DefaultRequestHeaders.Add(AccountTokenKey, _accountToken);
         }
      }

      public async Task<Stream> GetStream(string path, IEnumerable<Tuple<string, string>> parameters = null)
      {
         using (var client = new HttpClient())
         {
            SetHeaders(client);

            var uri = new Uri(FormatUrl(path, parameters));

            HttpResponseMessage result = Task.Run(async () => await client.GetAsync(uri)).Result;

            if (!result.IsSuccessStatusCode)
               return null;

            var content = await result.Content.ReadAsStreamAsync();
            return content;
         }
      }

      public async Task<T> Get<T>(string path, IEnumerable<Tuple<string, string>> parameters = null)
      {
         using (var client = new HttpClient())
         {
            SetHeaders(client);

            var uri = new Uri(FormatUrl(path, parameters));

            HttpResponseMessage result = Task.Run(async () => await client.GetAsync(uri)).Result;

            if (!result.IsSuccessStatusCode)
               return default(T);

            string content = await result.Content.ReadAsStringAsync();
            return JsonConvert.DeserializeObject<T>(content);
         }
      }

      public async Task<T> Get<T>(string hostnametoapi, string path, IEnumerable<Tuple<string, string>> parameters = null)
      {
         using (var client = new HttpClient())
         {
            SetHeaders(client);

            var uri = new Uri(FormatUrl(hostnametoapi, path, parameters));

            HttpResponseMessage result = Task.Run(async () => await client.GetAsync(uri)).Result;

            if (!result.IsSuccessStatusCode)
               return default(T);

            string content = await result.Content.ReadAsStringAsync();
            return JsonConvert.DeserializeObject<T>(content);
         }
      }

      public Stream PostAsJson(string relativeUrl, object data)
      {
         using (var client = new HttpClient())
         {
            SetHeaders(client);

            HttpResponseMessage result = Task.Run(async () => await client.PostAsJsonAsync(_baseUrl + relativeUrl, data)).Result;
            return result.Content.ReadAsStreamAsync().Result;
         }
      }

      public async Task<T> Post<T>(string relativeUrl, string data)
      {
         using (var client = new HttpClient())
         {
            SetHeaders(client);

            var content = new StringContent(data, Encoding.UTF8, "application/json");
            HttpResponseMessage result = await client.PostAsync(_baseUrl + relativeUrl, content);

            return await Deserialize<T>(result);
         }
      }

      public async Task<T> Multipart<T>(string relativeUrl, IDictionary<string, Stream> parts, HttpMethod method)
      {
         using (var client = new HttpClient())
         {
            SetHeaders(client);

            string formDataBoundary = string.Format("----------{0:N}", Guid.NewGuid());
            using (var content = new MultipartFormDataContent(formDataBoundary))
            {
               foreach (var part in parts)
               {
                  content.Add(new StreamContent(part.Value), part.Key);
               }

               if (method == HttpMethod.Post)
               {
                  using (HttpResponseMessage message = Task.Run(async () => await client.PostAsync(_baseUrl + relativeUrl, content)).Result)
                  {
                     return await Deserialize<T>(message);
                  }
               }

               if (method == HttpMethod.Put)
               {
                  using (HttpResponseMessage message = Task.Run(async () => await client.PutAsync(_baseUrl + relativeUrl, content)).Result)
                  {
                     return await Deserialize<T>(message);
                  }
               }

               throw new ArgumentException(string.Format("Could not submit multipart request on http method {0}", method));
            }
         }
      }

      private async Task<T> Deserialize<T>(HttpResponseMessage message)
      {
         string response = await message.Content.ReadAsStringAsync();

         if (!message.IsSuccessStatusCode)
         {
            return default(T);
         }

         return JsonConvert.DeserializeObject<T>(response);
      }

      private string FormatUrl(string path, IEnumerable<Tuple<string, string>> parameters)
      {
         return FormatUrl(_baseUrl, path, parameters);
      }

      private string FormatUrl(string baseUrl, string path, IEnumerable<Tuple<string, string>> parameters)
      {
         var buffer = new StringBuilder(1000);
         buffer.Append(baseUrl);
         buffer.Append(path);

         if (parameters != null)
         {
            bool first = false;
            foreach (var param in parameters)
            {
               if (!first)
               {
                  buffer.Append("?");
                  first = true;
               }
               else
               {
                  buffer.Append("&");
               }

               buffer.Append(param.Item1);
               buffer.Append("=");
               buffer.Append(param.Item2);
            }
         }
         return buffer.ToString();
      }

      public ResponseResult Login(string username, string password, [Optional] string accountId)
      {
         Guid clientNonce = Guid.NewGuid();

         ChallengeResult challengeResult =
            Task.Run(
               async () => await Post<ChallengeResult>("login/challenge", string.Format("{{ClientNonce:'{0}'}}", clientNonce))).Result;

         string clientHash = ComputeClientHash(challengeResult.ServerNonce, clientNonce, password);
         var responseContext = new ResponseContext
         {
            ClientHash = clientHash,
            ClientNonce = clientNonce,
            UserName = username
         };

         ResponseResult responseResult =
            Task.Run(async () => await Post<ResponseResult>("login/response", JsonConvert.SerializeObject(responseContext)))
               .Result;

         _sessionToken = responseResult.ApiSession.Token.ToString();

         if (accountId == null)
         {
            // account is is optional
            _accountToken = responseResult.Accounts[0].Id.ToString();
         }
         else
         {
            // check account id
            if (CheckAccountId(accountId, responseResult.Accounts))
            {
               // valid account id
               _accountToken = accountId;
            }
            else
            {
               // invalid account id
               throw new ArgumentException("Invalid Account Id");
            }
         }

         return responseResult;
      }

      // this function make sure account id is one of account id in accounts
      private bool CheckAccountId(string accountId, IList<ResponseResult.Account> accounts)
      {
         foreach (var account in accounts)
         {
            if (account.Id.Equals(new Guid(accountId)))
            {
               return true;
            }
         }
         return false;
      }

      public class ChallengeResult
      {
         public string ServerNonce { get; set; }
      }

      public class ResponseContext
      {
         public Guid ClientNonce { get; set; }
         public string ClientHash { get; set; }
         public string UserName { get; set; }
         public string ApplicationId { get; set; }
      }

      public class ResponseResult
      {
         public string ServerHash { get; set; }
         public Session ApiSession { get; set; }
         public IList<Account> Accounts { get; set; }
         public UserInfo User { get; set; }

         public class Account
         {
            public Guid Id { get; set; }
            public string Name { get; set; }
            public int AccessLevel { get; set; }
         }

         public class Session
         {
            public Guid Token { get; set; }
            public DateTime Expiry { get; set; }
         }

         public class UserInfo
         {
            public Guid Id { get; set; }
            public string FirstName { get; set; }
            public string LastName { get; set; }
            public string Email { get; set; }
            public bool IsFirstLogin { get; set; }
            public bool HasDocument { get; set; }
         }
      }

      private string ComputeClientHash(string serverNonce, Guid clientNonce, string password)
      {
         string data = clientNonce + serverNonce + password;

         var ue = new UTF8Encoding();
         var hsha = new HMACSHA1(ue.GetBytes(password));

         byte[] b = ue.GetBytes(data);
         byte[] h = hsha.ComputeHash(b);
         string computedHash = "";
         foreach (byte t in h)
            computedHash += string.Format("{0:X2}", t);

         return computedHash;
      }
   }
}